/*
 * Copyright (c) 2003 Chris Wareham
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Chris Wareham.
 * 4. The name of Chris Wareham may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CHRIS WAREHAM ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL CHRIS WAREHAM BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include "pkill.h"

#define OROBORUS_DIRECTORY "/usr/pkg/share/oroborus"
#define LINESZ 1024

enum {
	COLUMN_SELECTED,
	COLUMN_DESCRIPTION,
	NUM_COLUMNS
};

static gchar *read_rcfile(const gchar *);
static gint write_rcfile(const gchar *, const gchar *);
static gboolean file_copy(const gchar *, const gchar *);
static gint get_line(FILE *, gchar **);
static GtkTreeModel *create_model(const gchar *);
static void add_columns(GtkTreeView *);
static void selected_toggled(GtkCellRendererToggle *, gchar *, gpointer);
static gint delete_event(GtkWidget *, GdkEvent  *, gpointer);

/*
 * Display available themes for the Oroborus Window Manager.  If a new theme is
 * selected, then rewrite the configuration file and inform the window manager
 * via a signal.
 */
int
main(
	int argc,
	char *argv[])
{
	GtkWidget *window;
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *scrolled_window;
	GtkTreeModel *tree_model;
	GtkWidget *tree_view;
	gchar *current_theme;

	current_theme = read_rcfile("theme");

	gtk_init(&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Oroborus Theme Changer");
	gtk_container_set_border_width(GTK_CONTAINER(window), 8);
	gtk_window_set_default_size(GTK_WINDOW(window), 250, 250);

	g_signal_connect(G_OBJECT(window), "delete_event",
		G_CALLBACK(delete_event), NULL);

	vbox = gtk_vbox_new(FALSE, 8);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	label = gtk_label_new("Oroborus Theme Changer V1.0");
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window),
		GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
		GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);

	tree_model = create_model(current_theme);
	tree_view = gtk_tree_view_new_with_model(tree_model);
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree_view), TRUE);
	gtk_tree_view_set_search_column(GTK_TREE_VIEW(tree_view),
		COLUMN_DESCRIPTION);
	g_object_unref(tree_model);

	gtk_container_add(GTK_CONTAINER(scrolled_window), tree_view);

	add_columns(GTK_TREE_VIEW(tree_view));

	gtk_widget_show_all(window);

	gtk_main();

	return 0;
}

/*
 * Creates the configuration file if it's missing, then reads it to determine
 * the value for option "name".  Returns the option value if it's found, or
 * NULL if it isn't.
 */
static gchar *
read_rcfile(const gchar *name)
{
	const gchar *home;
	gchar *path, *line, *ptr;
	gboolean status;
	FILE *fp;
	gint namelen;

	home = g_getenv("HOME");
	if (home == NULL) {
		g_error("Unable to determine home directory\n");
		exit(1);
	}

	path = g_build_path("/", home, ".oroborusrc", NULL);

	status = g_file_test(path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR);
	if (status != TRUE) {
		g_warning("Creating %s\n", path);
		file_copy(OROBORUS_DIRECTORY "/defaults", path);
	}

	if ((fp = fopen(path, "r")) == NULL) {
		g_error("Unable to open %s: %s\n", path, strerror(errno));
		exit(1);
	}
	g_free(path);

	ptr = NULL;
	namelen = strlen(name);

	while (get_line(fp, &line) > 0) {
		if(strncmp(line, name, namelen) == 0 && line[namelen] == '=') {
			line[strlen(line) - 1] = '\0';
			if ((ptr = strrchr(line, '/')) == NULL || strlen(ptr) < 2) {
				g_warning("Bad theme entry in resource file: %s\n", line);
				g_free(line);
				break;
			}
			ptr = strdup(ptr + 1);
			g_free(line);
			break;
		}
		g_free(line);
	}

	fclose(fp);

	return ptr;
}

/*
 * Copy the contents of the file named "src" to the file named "dest".
 */
static gboolean
file_copy(
	const gchar *src,
	const gchar *dest)
{
	FILE *in, *out;
	gint n;
	gchar buf[1024];

	if ((in = fopen(src, "r")) == NULL)
		return FALSE;

	if ((out = fopen(dest, "w")) == NULL) {
		fclose(in);
		return FALSE;
	}
 
	while ((n = fread(buf, 1, sizeof(buf), in)) > 0)
		fwrite(buf, 1, n, out);

	fclose(in);
	fclose(out);

	return TRUE;
}

/*
 * Read from a file until a newline or end of file is encountered.  The buffer
 * pointed to by "*line" should be freed by the caller.
 */
static gint
get_line(FILE *fp, gchar **line) {
	gchar *ptr, buf[LINESZ];

	if ((*line = g_malloc(LINESZ)) == NULL)
		return -1;

	if ((fgets(*line, LINESZ, fp)) == NULL) {
		if (feof(fp)) {
			g_free(*line);
			return 0;
		}
		return -1;
	}

	if ((*line)[strlen(*line) - 1] != '\n') {
		while ((*line)[strlen(*line) - 1] != '\n') {
			if (feof(fp))
				break;
			fgets(buf, LINESZ, fp);
			if ((ptr = g_realloc(*line, strlen(*line) + LINESZ)) == NULL) {
				g_free(*line);
				return -1;
			}
			*line = ptr;
			ptr = *line + strlen(*line);
			strcpy(ptr, buf);
		}
	}

	return strlen(*line);
}

/*
 * Read entries in the Oroborus themes directory, adding them to a GTK+ list
 * store.
 */
static GtkTreeModel *
create_model(const gchar *current_theme)
{
	GtkListStore *list_store;
	GtkTreeIter tree_iter;
	GDir* dir;
	const gchar *name;
	gboolean selected;

	list_store = gtk_list_store_new(NUM_COLUMNS,
		G_TYPE_BOOLEAN,
		G_TYPE_STRING);

	dir = g_dir_open(OROBORUS_DIRECTORY "/themes", 0, NULL);
	if (dir != NULL) {
		while ((name = g_dir_read_name(dir)) != NULL) {
			if (strlen(name) < 1 || name[0] == '.')
				continue;
			if (current_theme != NULL && strcmp(name, current_theme) == 0)
				selected = TRUE;
			else
				selected = FALSE;
			gtk_list_store_append(list_store, &tree_iter);
			gtk_list_store_set(list_store, &tree_iter,
				COLUMN_SELECTED, selected,
				COLUMN_DESCRIPTION, name,
				-1);
		}
		g_dir_close(dir);
	}

	return GTK_TREE_MODEL(list_store);
}

/*
 * Create the GTK+ tree view with a column for the description and another with
 * a toggle to indicate if the theme is selected.
 */
static void
add_columns(GtkTreeView *tree_view)
{
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeModel *tree_model;

	tree_model = gtk_tree_view_get_model(tree_view);

	/* Column for selected toggles */
	renderer = gtk_cell_renderer_toggle_new();
	g_signal_connect(renderer, "toggled", G_CALLBACK(selected_toggled),
		tree_model);

	column = gtk_tree_view_column_new_with_attributes("Selected", renderer,
		"active", COLUMN_SELECTED, NULL);
	gtk_tree_view_column_set_sizing(GTK_TREE_VIEW_COLUMN(column),
		GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_column_set_fixed_width(GTK_TREE_VIEW_COLUMN(column), 70);
	gtk_tree_view_append_column(tree_view, column);

	/* Column for description */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes("Description", renderer,
		"text", COLUMN_DESCRIPTION, NULL);
	gtk_tree_view_column_set_sort_column_id(column, COLUMN_DESCRIPTION);
	gtk_tree_view_append_column(tree_view, column);
}

/*
 * Called when a toggle is clicked.  If a new theme has been selected, then
 * rewrite the configuration file and inform the window manager via a signal.
 */
static void
selected_toggled(
	GtkCellRendererToggle *cell,
	gchar *path_str,
	gpointer data)
{
	GtkTreeModel *tree_model;
	GtkTreePath *tree_path;
	GtkTreeIter selected_iter, tree_iter;
	gboolean valid, selected;
	gchar *description, *path;

	/* get toggled iterator */
	tree_model = GTK_TREE_MODEL(data);
	tree_path = gtk_tree_path_new_from_string(path_str);
	gtk_tree_model_get_iter(tree_model, &selected_iter, tree_path);
	gtk_tree_model_get(tree_model, &selected_iter,
		COLUMN_SELECTED, &selected,
		COLUMN_DESCRIPTION, &description,
		-1);
	gtk_tree_path_free(tree_path);

	/* If it's already selected, then just return */
	if (selected) {
		g_free(description);
		return;
	}

	/* clear the existing selection */
	valid = gtk_tree_model_get_iter_first(tree_model, &tree_iter);
	while (valid) {
		gtk_tree_model_get(tree_model, &tree_iter, 
			COLUMN_SELECTED, &selected,
			-1);
		if (selected) {
			selected = FALSE;
			gtk_list_store_set(GTK_LIST_STORE(tree_model), &tree_iter,
				COLUMN_SELECTED, selected,
				-1);
			break;
		}
		valid = gtk_tree_model_iter_next(tree_model, &tree_iter);
    }

	/* set new value */
	selected = TRUE;
	gtk_list_store_set(GTK_LIST_STORE(tree_model), &selected_iter,
		COLUMN_SELECTED, selected,
		-1);

	/* write new theme selection to resource file */
	path = g_build_path("/", OROBORUS_DIRECTORY, "themes", description, NULL);
	g_free(description);
	write_rcfile("theme", path);
	g_free(path);

	/* inform window manager of configuration change */
	pkill("oroborus", SIGHUP);
}

/*
 * Rewrites the configuration file, updating the option "name" with "value".
 * Returns 0 if the option is found, or 1 if it isn't.
 */
static int
write_rcfile(const char *name, const char *value)
{
	const gchar *home;
	gchar *in_path, *out_path, *line;
	FILE *in, *out;
	gint namelen, found;

	home = g_getenv("HOME");
	if (home == NULL) {
		g_error("Unable to determine home directory\n");
		exit(1);
	}

	in_path = g_build_path("/", home, ".oroborusrc", NULL);

	if ((in = fopen(in_path, "r")) == NULL) {
		g_error("Unable to open %s: %s\n", in_path, strerror(errno));
		exit(1);
	}

	out_path = g_build_path("/", home, ".oroborusrc.tmp", NULL);

	if ((out = fopen(out_path, "w")) == NULL) {
		g_error("Unable to open %s: %s\n", out_path, strerror(errno));
		exit(1);
	}

	found = 1;
	namelen = strlen(name);
	
	while (get_line(in, &line) > 0) {
		if(strncmp(line, name, namelen) == 0 && line[namelen] == '=') {
			found = 0;
			fprintf(out, "%s=%s\n", name, value);
		} else
			fprintf(out, "%s", line);
		g_free(line);
	}

	fclose(in);
	fclose(out);

	rename(out_path, in_path);

	g_free(in_path);
	g_free(out_path);

	return found;
}

/*
 * Called when the close button is clicked on the window frame.
 */
static gint
delete_event(
	GtkWidget *widget,
	GdkEvent *event,
	gpointer data)
{
	gtk_main_quit();
	return FALSE;
}
